/* PSPP - a program for statistical analysis.
   Copyright (C) 2012 Free Software Foundation, Inc.

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#ifndef OUTPUT_CHARTS_SL_PLOT_H
#define OUTPUT_CHARTS_SL_PLOT_H 1

#include "output/chart-item.h"

struct datum 
{
  double x;
  double y;
};

struct spreadlevel_plot_chart
  {
    struct chart_item chart_item;

    /* Calculated parameters */
    double y_lower, y_upper;
    double x_lower, x_upper;

    double tx_pwr;

    size_t n_data;
    struct datum *data;
  };

struct chart_item *spreadlevel_plot_create (const char *label, double slp);

void spreadlevel_plot_add (struct chart_item *, double spread, double level);



/* This boilerplate for spreadlevel_plot_chart, a subclass of chart_item, was
   autogenerated by mk-class-boilerplate. */

#include <assert.h>
#include "libpspp/cast.h"

extern const struct chart_item_class spreadlevel_plot_chart_class;

/* Returns true if SUPER is a spreadlevel_plot_chart, otherwise false. */
static inline bool
is_spreadlevel_plot_chart (const struct chart_item *super)
{
  return super->class == &spreadlevel_plot_chart_class;
}

/* Returns SUPER converted to spreadlevel_plot_chart.  SUPER must be a spreadlevel_plot_chart, as
   reported by is_spreadlevel_plot_chart. */
static inline struct spreadlevel_plot_chart *
to_spreadlevel_plot_chart (const struct chart_item *super)
{
  assert (is_spreadlevel_plot_chart (super));
  return UP_CAST (super, struct spreadlevel_plot_chart, chart_item);
}

/* Returns INSTANCE converted to chart_item. */
static inline struct chart_item *
spreadlevel_plot_chart_super (const struct spreadlevel_plot_chart *instance)
{
  return CONST_CAST (struct chart_item *, &instance->chart_item);
}

/* Increments INSTANCE's reference count and returns INSTANCE. */
static inline struct spreadlevel_plot_chart *
spreadlevel_plot_chart_ref (const struct spreadlevel_plot_chart *instance)
{
  return to_spreadlevel_plot_chart (chart_item_ref (&instance->chart_item));
}

/* Decrements INSTANCE's reference count, then destroys INSTANCE if
   the reference count is now zero. */
static inline void
spreadlevel_plot_chart_unref (struct spreadlevel_plot_chart *instance)
{
  chart_item_unref (&instance->chart_item);
}

/* Returns true if INSTANCE's reference count is greater than 1,
   false otherwise. */
static inline bool
spreadlevel_plot_chart_is_shared (const struct spreadlevel_plot_chart *instance)
{
  return chart_item_is_shared (&instance->chart_item);
}

static inline void
spreadlevel_plot_chart_submit (struct spreadlevel_plot_chart *instance)
{
  chart_item_submit (&instance->chart_item);
}

#endif 
