/* PSPP - a program for statistical analysis.
   Copyright (C) 1997-9, 2000, 2010 Free Software Foundation, Inc.

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#if !version_h
#define version_h 1

/* "A.B.C" */
extern const char bare_version[];

/* "GNU PSPP A.B.C" */
extern const char version[];

/* "GNU PSPP version A.B (date), Copyright (C) XXXX Free Software
   Foundation, Inc." */
extern const char stat_version[];

/* Canonical name of host system type. */
extern const char host_system[];

/* Canonical name of build system type. */
extern const char build_system[];

/* Locale directory. */
extern const char locale_dir[];

/* From the AUTHORS file */
extern const char *const authors[];

#endif /* !version_h */
