/* PSPP - a program for statistical analysis.
   Copyright (C) 1997-9, 2000, 2006, 2007, 2008 Free Software Foundation, Inc.

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#ifndef LANGUAGE_DATA_IO_TRIM_H
#define LANGUAGE_DATA_IO_TRIM_H

#include <stdbool.h>

struct lexer;
struct dictionary;
bool parse_dict_trim (struct lexer *, struct dictionary *);
bool parse_dict_rename (struct lexer *, struct dictionary *);
bool parse_dict_drop (struct lexer *, struct dictionary *);
bool parse_dict_keep (struct lexer *, struct dictionary *);

#endif /* trim.c */
