/* PSPP - a program for statistical analysis.
   Copyright (C) 1997-9, 2000, 2010, 2011 Free Software Foundation, Inc.

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#ifndef LANGUAGE_DATA_IO_FILE_HANDLE_H
#define LANGUAGE_DATA_IO_FILE_HANDLE_H 1

/* Parsing file handles. */

#include <stdbool.h>
#include <stddef.h>
#include "data/file-handle-def.h"

struct lexer;
struct session;

struct file_handle *fh_parse (struct lexer *, enum fh_referent,
                              struct session *);

#endif  /* language/data-io/file-handle.h */
